@@ -7,14 +7,16 @@
 //! `zerovec` enable vectors of multibyte types to be backed by a byte array, abstracting away
 //! issues including memory alignment and endianness.
 //!
-//! This crate has two main types:
+//! This crate has three main types:
 //!
-//! - `ZeroVec<T>` for fixed-width types like `u32`
-//! - `VarZeroVec<T>` for variable-width types like `str`
+//! - [`ZeroVec<T>`](ZeroVec) for fixed-width types like `u32`
+//! - [`VarZeroVec<T>`](VarZeroVec) for variable-width types like `str`
+//! - [`ZeroMap<K, V>`](ZeroMap) to map from `K` to `V`
 //!
-//! Both are intended as drop-in replacements for `Vec<T>` in Serde structs serialized with a
-//! format supporting a borrowed byte buffer, like Bincode. Clients upgrading from Vec to ZeroVec
-//! or VarZeroVec benefit from zero heap allocations when deserializing read-only data.
+//! The first two are intended as drop-in replacements for `Vec<T>` in Serde structs serialized
+//! with a format supporting a borrowed byte buffer, like Bincode. The third is indended as a
+//! replacement for `HashMap` or `LiteMap`. Clients upgrading to `ZeroVec`, `VarZeroVec`, or
+//! `ZeroMap` benefit from zero heap allocations when deserializing read-only data.
 //!
 //! This crate has two optional features: `serde` and `yoke`. `serde` allows serializing and deserializing
 //! `zerovec`'s abstractions via [`serde`](https://docs.rs/serde), and `yoke` enables implementations of `Yokeable`
@@ -32,13 +34,24 @@
 //! | Deserialize vec of 100 `u32` | 233.18 ns | 14.120 ns |
 //! | Compute sum of vec of 100 `u32` (read every element) | 8.7472 ns | 10.775 ns |
 //! | Binary search vec of 1000 `u32` 50 times | 442.80 ns | 472.51 ns |
-//! | Deserialize vec of 100 strings | 7.3740 us\* | 1.4495 us |
+//! | Deserialize vec of 100 strings | 7.3740 μs\* | 1.4495 μs |
 //! | Count chars in vec of 100 strings (read every element) | 747.50 ns | 955.28 ns |
 //! | Binary search vec of 500 strings 10 times | 466.09 ns | 790.33 ns |
 //!
-//! \* *This result is reported for `Vec<String>`. However, Serde also supports deserializing to `Vec<&str>`; this gives 1.8420 us, much faster than `Vec<String>` but a bit slower than `zerovec`.*
+//! \* *This result is reported for `Vec<String>`. However, Serde also supports deserializing to `Vec<&str>`; this gives 1.8420 μs, much faster than `Vec<String>` but a bit slower than `zerovec`.*
+//!
+//! | Operation | `HashMap<K,V>`  | `LiteMap<K,V>` | `ZeroMap<K,V>` |
+//! |---|---|---|---|
+//! | Deserialize a small map | 2.72 μs | 1.28 μs | 480 ns |
+//! | Deserialize a large map | 50.5 ms | 18.3 ms | 3.74 ms |
+//! | Look up from a small deserialized map | 49 ns | 42 ns | 54 ns |
+//! | Look up from a large deserialized map | 51 ns | 155 ns | 213 ns |
+//!
+//! Small = 16 elements, large = 131,072 elements. Maps contain `<String, String>`.
 //!
 //! The benches used to generate the above table can be found in the `benches` directory in the project repository.
+//! `zeromap` benches are named by convention, e.g. `zeromap/deserialize/small`, `zeromap/lookup/large`. The type
+//! is appended for baseline comparisons, e.g. `zeromap/lookup/small/hashmap`.
 //!
 //! # Features
 //!
@@ -58,12 +71,12 @@
 //!     #[serde(borrow)]
 //!     nums: ZeroVec<'data, u32>,
 //!     #[serde(borrow)]
-//!     strs: VarZeroVec<'data, String>,
+//!     strs: VarZeroVec<'data, str>,
 //! }
 //!
 //! let data = DataStruct {
 //!     nums: ZeroVec::from_slice(&[211, 281, 421, 461]),
-//!     strs: VarZeroVec::from(vec!["hello".to_string(), "world".to_string()]),
+//!     strs: VarZeroVec::from(&["hello", "world"] as &[_]),
 //! };
 //! let bincode_bytes = bincode::serialize(&data)
 //!     .expect("Serialization should be successful");
@@ -77,11 +90,18 @@
 //! # } // feature = "serde"
 //! ```
 
+#![cfg_attr(not(any(test, feature = "std")), no_std)]
+// this crate does a lot of nuanced lifetime manipulation, being explicit
+// is better here.
+#![allow(clippy::needless_lifetimes)]
+
+extern crate alloc;
+
 pub mod map;
 #[cfg(test)]
 pub mod samples;
 pub mod ule;
-mod varzerovec;
+pub mod varzerovec;
 mod zerovec;
 
 #[cfg(feature = "yoke")]
