@@ -6,7 +6,7 @@
 
 use crate::ule::AsULE;
 use crate::ZeroVec;
-use std::cmp::Ordering;
+use core::cmp::Ordering;
 
 mod kv;
 #[cfg(feature = "serde")]
@@ -17,15 +17,40 @@ pub use kv::ZeroMapKV;
 pub use vecs::ZeroVecLike;
 
 /// A zero-copy map datastructure, built on sorted binary-searchable [`ZeroVec`]
-/// and [`VarZeroVec`](crate::VarZeroVec).
+/// and [`VarZeroVec`].
 ///
-/// This type, like [`ZeroVec`] and [`VarZeroVec`](crate::VarZeroVec), is able to zero-copy
+/// This type, like [`ZeroVec`] and [`VarZeroVec`], is able to zero-copy
 /// deserialize from appropriately formatted byte buffers. It is internally copy-on-write, so it can be mutated
 /// afterwards as necessary.
+///
+/// Internally, a `ZeroMap` is a zero-copy vector for keys paired with a zero-copy vector for
+/// values, sorted by the keys. Therefore, all types used in `ZeroMap` need to work with either
+/// [`ZeroVec`] or [`VarZeroVec`].
+///
+/// # Examples
+///
+/// ```
+/// use zerovec::ZeroMap;
+///
+/// // Example byte buffer representing the map { 1: "one" }
+/// let BINCODE_BYTES: &[u8; 31] = &[
+///     4, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 11, 0, 0, 0, 0, 0, 0, 0,
+///     1, 0, 0, 0, 0, 0, 0, 0, 111, 110, 101
+/// ];
+///
+/// // Deserializing to ZeroMap requires no heap allocations.
+/// let zero_map: ZeroMap<u32, str> = bincode::deserialize(BINCODE_BYTES)
+///     .expect("Should deserialize successfully");
+/// assert_eq!(zero_map.get(&1), Some("one"));
+/// ```
+///
+/// [`VarZeroVec`]: crate::VarZeroVec
 pub struct ZeroMap<'a, K, V>
 where
     K: ZeroMapKV<'a>,
     V: ZeroMapKV<'a>,
+    K: ?Sized,
+    V: ?Sized,
 {
     pub(crate) keys: K::Container,
     pub(crate) values: V::Container,
@@ -35,6 +60,8 @@ impl<'a, K, V> Default for ZeroMap<'a, K, V>
 where
     K: ZeroMapKV<'a>,
     V: ZeroMapKV<'a>,
+    K: ?Sized,
+    V: ?Sized,
 {
     fn default() -> Self {
         Self {
@@ -48,6 +75,8 @@ impl<'a, K, V> ZeroMap<'a, K, V>
 where
     K: ZeroMapKV<'a>,
     V: ZeroMapKV<'a>,
+    K: ?Sized,
+    V: ?Sized,
 {
     /// Construct a new [`ZeroMap`]
     pub fn new() -> Self {
@@ -93,8 +122,8 @@ where
     /// use zerovec::ZeroMap;
     ///
     /// let mut map = ZeroMap::new();
-    /// map.insert(1, "one".to_owned());
-    /// map.insert(2, "two".to_owned());
+    /// map.insert(&1, "one");
+    /// map.insert(&2, "two");
     /// assert_eq!(map.get(&1), Some("one"));
     /// assert_eq!(map.get(&3), None);
     /// ```
@@ -109,8 +138,8 @@ where
     /// use zerovec::ZeroMap;
     ///
     /// let mut map = ZeroMap::new();
-    /// map.insert(1, "one".to_owned());
-    /// map.insert(2, "two".to_owned());
+    /// map.insert(&1, "one");
+    /// map.insert(&2, "two");
     /// assert_eq!(map.contains_key(&1), true);
     /// assert_eq!(map.contains_key(&3), false);
     /// ```
@@ -124,12 +153,12 @@ where
     /// use zerovec::ZeroMap;
     ///
     /// let mut map = ZeroMap::new();
-    /// map.insert(1, "one".to_owned());
-    /// map.insert(2, "two".to_owned());
+    /// map.insert(&1, "one");
+    /// map.insert(&2, "two");
     /// assert_eq!(map.get(&1), Some("one"));
     /// assert_eq!(map.get(&3), None);
     /// ```
-    pub fn insert(&mut self, key: K, value: V) -> Option<V> {
+    pub fn insert(&mut self, key: &K, value: &V) -> Option<V::OwnedType> {
         let key_needle = key.as_needle();
         match self.keys.binary_search(key_needle) {
             Ok(index) => Some(self.values.replace(index, value)),
@@ -147,12 +176,12 @@ where
     /// use zerovec::ZeroMap;
     ///
     /// let mut map = ZeroMap::new();
-    /// map.insert(1, "one".to_owned());
-    /// map.insert(2, "two".to_owned());
-    /// assert_eq!(map.remove(&1), Some("one".to_owned()));
+    /// map.insert(&1, "one");
+    /// map.insert(&2, "two");
+    /// assert_eq!(map.remove(&1), Some("one".to_owned().into_boxed_str()));
     /// assert_eq!(map.get(&1), None);
     /// ```
-    pub fn remove(&mut self, key: &K::NeedleType) -> Option<V> {
+    pub fn remove(&mut self, key: &K::NeedleType) -> Option<V::OwnedType> {
         let idx = self.keys.binary_search(key).ok()?;
         self.keys.remove(idx);
         Some(self.values.remove(idx))
@@ -165,13 +194,13 @@ where
     /// use zerovec::ZeroMap;
     ///
     /// let mut map = ZeroMap::new();
-    /// assert!(map.try_append(1, "uno".to_owned()).is_none());
-    /// assert!(map.try_append(3, "tres".to_owned()).is_none());
+    /// assert!(map.try_append(&1, "uno").is_none());
+    /// assert!(map.try_append(&3, "tres").is_none());
     ///
-    /// let unsuccessful = map.try_append(3, "tres-updated".to_owned());
+    /// let unsuccessful = map.try_append(&3, "tres-updated");
     /// assert!(unsuccessful.is_some(), "append duplicate of last key");
     ///
-    /// let unsuccessful = map.try_append(2, "dos".to_owned());
+    /// let unsuccessful = map.try_append(&2, "dos");
     /// assert!(unsuccessful.is_some(), "append out of order");
     ///
     /// assert_eq!(map.get(&1), Some("uno"));
@@ -183,7 +212,7 @@ where
     /// assert_eq!(map.get(&2), None);
     /// ```
     #[must_use]
-    pub fn try_append(&mut self, key: K, value: V) -> Option<(K, V)> {
+    pub fn try_append<'b>(&mut self, key: &'b K, value: &'b V) -> Option<(&'b K, &'b V)> {
         if self.keys.len() != 0 {
             if let Some(last) = self.keys.get(self.keys.len() - 1) {
                 if key.cmp_get(last) != Ordering::Greater {
