@@ -26,21 +26,32 @@ macro_rules! impl_byte_slice_size {
                 &self.0
             }
         }
-        impl ULE for PlainOldULE<$size> {
-            type Error = std::convert::Infallible;
+        // This is safe to implement because from_byte_slice_unchecked returns
+        // the same value as parse_byte_slice
+        unsafe impl ULE for PlainOldULE<$size> {
+            type Error = ULEError<core::convert::Infallible>;
+
             #[inline]
-            fn parse_byte_slice(bytes: &[u8]) -> Result<&[Self], Self::Error> {
-                let data = bytes.as_ptr();
+            fn validate_byte_slice(bytes: &[u8]) -> Result<(), Self::Error> {
+                if bytes.len() % $size == 0 {
+                    // Safe because Self is transparent over [u8; $size]
+                    Ok(())
+                } else {
+                    Err(ULEError::InvalidLength {
+                        ty: concat!("PlainOldULE<", stringify!($size), ">"),
+                        len: bytes.len(),
+                    })
+                }
+            }
+        }
+
+        impl PlainOldULE<$size> {
+            #[inline]
+            pub fn from_byte_slice_unchecked_mut(bytes: &mut [u8]) -> &mut [Self] {
+                let data = bytes.as_mut_ptr();
                 let len = bytes.len() / $size;
                 // Safe because Self is transparent over [u8; $size]
-                Ok(unsafe { std::slice::from_raw_parts(data as *const Self, len) })
-            }
-            #[inline]
-            fn as_byte_slice(slice: &[Self]) -> &[u8] {
-                let data = slice.as_ptr();
-                let len = slice.len() * $size;
-                // Safe because Self is transparent over [u8; $size]
-                unsafe { std::slice::from_raw_parts(data as *const u8, len) }
+                unsafe { core::slice::from_raw_parts_mut(data as *mut Self, len) }
             }
         }
     };
@@ -57,11 +68,11 @@ macro_rules! impl_byte_slice_type {
         impl AsULE for $type {
             type ULE = PlainOldULE<$size>;
             #[inline]
-            fn as_unaligned(&self) -> Self::ULE {
+            fn as_unaligned(self) -> Self::ULE {
                 PlainOldULE(self.to_le_bytes())
             }
             #[inline]
-            fn from_unaligned(unaligned: &Self::ULE) -> Self {
+            fn from_unaligned(unaligned: Self::ULE) -> Self {
                 <$type>::from_le_bytes(unaligned.0)
             }
         }
@@ -71,20 +82,54 @@ macro_rules! impl_byte_slice_type {
     };
 }
 
-impl_byte_slice_size!(1);
 impl_byte_slice_size!(2);
 impl_byte_slice_size!(4);
 impl_byte_slice_size!(8);
 impl_byte_slice_size!(16);
 
-impl_byte_slice_type!(u8, 1);
 impl_byte_slice_type!(u16, 2);
 impl_byte_slice_type!(u32, 4);
 impl_byte_slice_type!(u64, 8);
 impl_byte_slice_type!(u128, 16);
 
-impl_byte_slice_type!(i8, 1);
 impl_byte_slice_type!(i16, 2);
 impl_byte_slice_type!(i32, 4);
 impl_byte_slice_type!(i64, 8);
 impl_byte_slice_type!(i128, 16);
+
+// This is safe to implement because from_byte_slice_unchecked returns
+// the same value as parse_byte_slice
+unsafe impl ULE for u8 {
+    type Error = core::convert::Infallible;
+    #[inline]
+    fn validate_byte_slice(_bytes: &[u8]) -> Result<(), Self::Error> {
+        Ok(())
+    }
+    #[inline]
+    fn parse_byte_slice(bytes: &[u8]) -> Result<&[Self], Self::Error> {
+        Ok(bytes)
+    }
+    #[inline]
+    unsafe fn from_byte_slice_unchecked(bytes: &[u8]) -> &[Self] {
+        bytes
+    }
+    #[inline]
+    fn as_byte_slice(slice: &[Self]) -> &[u8] {
+        slice
+    }
+}
+
+impl AsULE for u8 {
+    type ULE = Self;
+    #[inline]
+    fn as_unaligned(self) -> Self::ULE {
+        self
+    }
+    #[inline]
+    fn from_unaligned(unaligned: Self::ULE) -> Self {
+        unaligned
+    }
+}
+
+// EqULE is true because u8 is its own ULE.
+unsafe impl EqULE for u8 {}
