@@ -3,23 +3,19 @@
 // (online at: https://github.com/unicode-org/icu4x/blob/main/LICENSE ).
 
 use crate::ule::*;
-use std::str;
+use core::str;
 
-impl AsVarULE for String {
-    type VarULE = str;
-    #[inline]
-    fn as_unaligned(&self) -> &str {
-        self
-    }
-    #[inline]
-    fn from_unaligned(unaligned: &str) -> Self {
-        unaligned.into()
-    }
-}
-
-impl VarULE for str {
+// This is safe to implement because from_byte_slice_unchecked returns
+// the same value as parse_byte_slice
+unsafe impl VarULE for str {
     type Error = str::Utf8Error;
 
+    #[inline]
+    fn validate_byte_slice(bytes: &[u8]) -> Result<(), Self::Error> {
+        str::from_utf8(bytes)?;
+        Ok(())
+    }
+
     #[inline]
     fn parse_byte_slice(bytes: &[u8]) -> Result<&Self, Self::Error> {
         str::from_utf8(bytes)
