@@ -6,7 +6,9 @@
 mod serde;
 
 use crate::ule::*;
-use std::fmt;
+use alloc::vec::Vec;
+use core::fmt;
+use core::iter::FromIterator;
 
 /// A zero-copy vector for fixed-width types.
 ///
@@ -44,7 +46,7 @@ use std::fmt;
 /// let nums: &[u16] = &[211, 281, 421, 461];
 ///
 /// // Conversion from &[u8] to &[u16::ULE] is infallible.
-/// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+/// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
 ///
 /// assert!(matches!(zerovec, ZeroVec::Borrowed(_)));
 /// assert_eq!(zerovec.get(2), Some(421));
@@ -58,7 +60,31 @@ pub enum ZeroVec<'a, T>
 where
     T: AsULE + ?Sized,
 {
+    /// An owned `ZeroVec<T>`. This will typically be constructed by [`ZeroVec::clone_from_slice()`]
+    /// or by calling [`ZeroVec::to_mut()`]/[`ZeroVec::for_each_mut()`]/etc on [`ZeroVec::Borrowed`].
     Owned(Vec<T::ULE>),
+
+    /// A borrowed `ZeroVec<T>`. This will typically be constructed by [`ZeroVec::parse_byte_slice()`],
+    /// [`ZeroVec::from_slice()`], or deserializers capable of doing zero-copy deserialization.
+    ///
+    /// If you already have a slice of `[T::ULE]`s, you can directly construct one of these.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use zerovec::ZeroVec;
+    /// use zerovec::ule::*;
+    ///
+    /// // The little-endian bytes correspond to the numbers on the following line.
+    /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
+    /// let nums: &[PlainOldULE<2>] = &[211_u16.as_unaligned(), 281_u16.as_unaligned(),
+    ///                                 421_u16.as_unaligned(), 461_u16.as_unaligned()];
+    ///
+    /// let zerovec = ZeroVec::<u16>::Borrowed(nums);
+    ///
+    /// assert!(matches!(zerovec, ZeroVec::Borrowed(_)));
+    /// assert_eq!(bytes, zerovec.as_bytes());
+    /// ```
     Borrowed(&'a [T::ULE]),
 }
 
@@ -74,9 +100,11 @@ where
     }
 }
 
+impl<T> Eq for ZeroVec<'_, T> where T: AsULE + Eq + ?Sized {}
+
 impl<'a, 'b, T> PartialEq<ZeroVec<'b, T>> for ZeroVec<'a, T>
 where
-    T: AsULE + Copy + PartialEq,
+    T: AsULE + PartialEq + ?Sized,
 {
     #[inline]
     fn eq(&self, other: &ZeroVec<'b, T>) -> bool {
@@ -87,7 +115,7 @@ where
 
 impl<T> PartialEq<&[T]> for ZeroVec<'_, T>
 where
-    T: AsULE + Copy + PartialEq,
+    T: AsULE + PartialEq + ?Sized,
 {
     #[inline]
     fn eq(&self, other: &&[T]) -> bool {
@@ -117,12 +145,12 @@ where
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert!(matches!(zerovec, ZeroVec::Borrowed(_)));
     /// assert_eq!(zerovec.get(2), Some(421));
     /// ```
-    pub fn try_from_bytes(bytes: &'a [u8]) -> Result<Self, <<T as AsULE>::ULE as ULE>::Error> {
+    pub fn parse_byte_slice(bytes: &'a [u8]) -> Result<Self, <<T as AsULE>::ULE as ULE>::Error> {
         let slice: &'a [T::ULE] = T::ULE::parse_byte_slice(bytes)?;
         Ok(Self::Borrowed(slice))
     }
@@ -168,7 +196,7 @@ where
     /// use zerovec::ule::AsULE;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert_eq!(4, zerovec.len());
     /// assert_eq!(
@@ -189,10 +217,10 @@ where
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     /// assert!(!zerovec.is_empty());
     ///
-    /// let emptyvec: ZeroVec<u16> = ZeroVec::try_from_bytes(&[]).expect("infallible");
+    /// let emptyvec: ZeroVec<u16> = ZeroVec::parse_byte_slice(&[]).expect("infallible");
     /// assert!(emptyvec.is_empty());
     /// ```
     #[inline]
@@ -225,7 +253,7 @@ where
     /// ```
     #[inline]
     pub fn clone_from_slice(other: &[T]) -> Self {
-        Self::Owned(other.iter().map(T::as_unaligned).collect())
+        Self::Owned(other.iter().copied().map(T::as_unaligned).collect())
     }
 
     /// Creates a `Vec<T>` from a `ZeroVec<T>`.
@@ -242,7 +270,11 @@ where
     /// ```
     #[inline]
     pub fn to_vec(&self) -> Vec<T> {
-        self.as_slice().iter().map(T::from_unaligned).collect()
+        self.as_slice()
+            .iter()
+            .copied()
+            .map(T::from_unaligned)
+            .collect()
     }
 }
 
@@ -301,26 +333,24 @@ where
 
 impl<T> ZeroVec<'_, T>
 where
-    T: AsULE + Copy + ?Sized,
+    T: AsULE,
 {
     /// Gets the element at the specified index. Returns None if out of range.
     ///
-    /// The element is returned by value, so `T` must implement `Copy`.
-    ///
     /// # Example
     ///
     /// ```
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert_eq!(zerovec.get(2), Some(421));
     /// assert_eq!(zerovec.get(4), None);
     /// ```
     #[inline]
     pub fn get(&self, index: usize) -> Option<T> {
-        self.as_slice().get(index).map(T::from_unaligned)
+        self.as_slice().get(index).copied().map(T::from_unaligned)
     }
 
     pub(crate) fn get_ule_ref(&self, index: usize) -> Option<&T::ULE> {
@@ -329,53 +359,47 @@ where
 
     /// Gets the first element. Returns None if empty.
     ///
-    /// The element is returned by value, so `T` must implement `Copy`.
-    ///
     /// # Example
     ///
     /// ```
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert_eq!(zerovec.first(), Some(211));
     /// ```
     #[inline]
     pub fn first(&self) -> Option<T> {
-        self.as_slice().first().map(T::from_unaligned)
+        self.as_slice().first().copied().map(T::from_unaligned)
     }
 
     /// Gets the last element. Returns None if empty.
     ///
-    /// The element is returned by value, so `T` must implement `Copy`.
-    ///
     /// # Example
     ///
     /// ```
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert_eq!(zerovec.last(), Some(461));
     /// ```
     #[inline]
     pub fn last(&self) -> Option<T> {
-        self.as_slice().last().map(T::from_unaligned)
+        self.as_slice().last().copied().map(T::from_unaligned)
     }
 
     /// Gets an iterator over the elements.
     ///
-    /// The elements are returned by value, so `T` must implement `Copy`.
-    ///
     /// # Example
     ///
     /// ```
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     /// let mut it = zerovec.iter();
     ///
     /// assert_eq!(it.next(), Some(211));
@@ -386,7 +410,68 @@ where
     /// ```
     #[inline]
     pub fn iter(&self) -> impl Iterator<Item = T> + '_ {
-        self.as_slice().iter().map(T::from_unaligned)
+        self.as_slice().iter().copied().map(T::from_unaligned)
+    }
+
+    /// Mutates each element according to a given function, meant to be
+    /// a more convenient version of [`ZeroVec::iter_mut()`] which serves fewer use cases.
+    ///
+    /// This will convert the ZeroVec into an owned ZeroVec if not already the case.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use zerovec::ZeroVec;
+    /// use zerovec::ule::AsULE;
+    ///
+    /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
+    /// let mut zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
+    ///
+    /// zerovec.for_each_mut(|item| *item += 1);
+    ///
+    /// assert_eq!(zerovec.to_vec(), &[212, 282, 422, 462]);
+    /// assert!(matches!(zerovec, ZeroVec::Owned(_)));
+    /// ```
+    #[inline]
+    pub fn for_each_mut(&mut self, mut f: impl FnMut(&mut T)) {
+        self.to_mut().iter_mut().for_each(|item| {
+            let mut aligned = T::from_unaligned(*item);
+            f(&mut aligned);
+            *item = aligned.as_unaligned()
+        });
+    }
+
+    /// Same as [`ZeroVec::for_each_mut()`], but bubbles up errors.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use zerovec::ZeroVec;
+    /// use zerovec::ule::AsULE;
+    ///
+    /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
+    /// let mut zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
+    ///
+    /// zerovec.try_for_each_mut(|item| {
+    ///     *item = item.checked_add(1).ok_or(())?;
+    ///     Ok(())   
+    /// })?;
+    ///
+    /// assert_eq!(zerovec.to_vec(), &[212, 282, 422, 462]);
+    /// assert!(matches!(zerovec, ZeroVec::Owned(_)));
+    /// # Ok::<(), ()>(())
+    /// ```
+    #[inline]
+    pub fn try_for_each_mut<E>(
+        &mut self,
+        mut f: impl FnMut(&mut T) -> Result<(), E>,
+    ) -> Result<(), E> {
+        self.to_mut().iter_mut().try_for_each(|item| {
+            let mut aligned = T::from_unaligned(*item);
+            f(&mut aligned)?;
+            *item = aligned.as_unaligned();
+            Ok(())
+        })
     }
 
     /// Converts a borrowed ZeroVec to an owned ZeroVec. No-op if already owned.
@@ -397,7 +482,7 @@ where
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     /// assert!(matches!(zerovec, ZeroVec::Borrowed(_)));
     ///
     /// let owned = zerovec.into_owned();
@@ -407,13 +492,14 @@ where
         match self {
             Self::Owned(vec) => ZeroVec::Owned(vec),
             Self::Borrowed(_) => {
-                let vec: Vec<T::ULE> = self.iter().map(|ule| T::as_unaligned(&ule)).collect();
+                let vec: Vec<T::ULE> = self.iter().map(T::as_unaligned).collect();
                 ZeroVec::Owned(vec)
             }
         }
     }
 
-    /// Allows the ZeroVec to be mutated by converting it to an owned variant
+    /// Allows the ZeroVec to be mutated by converting it to an owned variant, and producing
+    /// a mutable vector of ULEs.
     ///
     /// # Example
     ///
@@ -422,24 +508,21 @@ where
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let mut zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let mut zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     /// assert!(matches!(zerovec, ZeroVec::Borrowed(_)));
     ///
-    /// zerovec.make_mut().push(12_u16.as_unaligned());
+    /// zerovec.to_mut().push(12_u16.as_unaligned());
     /// assert!(matches!(zerovec, ZeroVec::Owned(_)));
     /// ```
-    //
-    // This function is crate-public for now since we don't yet want to stabilize
-    // the internal implementation details
-    pub(crate) fn make_mut(&mut self) -> &mut Vec<T::ULE> {
+    pub fn to_mut(&mut self) -> &mut Vec<T::ULE> {
         match self {
             ZeroVec::Owned(ref mut vec) => vec,
             ZeroVec::Borrowed(_) => {
-                let vec: Vec<T::ULE> = self.iter().map(|ule| T::as_unaligned(&ule)).collect();
+                let vec: Vec<T::ULE> = self.iter().map(T::as_unaligned).collect();
                 let new_self = ZeroVec::Owned(vec);
                 *self = new_self;
                 // recursion is limited since we are guaranteed to hit the Owned branch
-                self.make_mut()
+                self.to_mut()
             }
         }
     }
@@ -458,7 +541,7 @@ where
     /// use zerovec::ZeroVec;
     ///
     /// let bytes: &[u8] = &[0xD3, 0x00, 0x19, 0x01, 0xA5, 0x01, 0xCD, 0x01];
-    /// let zerovec: ZeroVec<u16> = ZeroVec::try_from_bytes(bytes).expect("infallible");
+    /// let zerovec: ZeroVec<u16> = ZeroVec::parse_byte_slice(bytes).expect("infallible");
     ///
     /// assert_eq!(zerovec.binary_search(&281), Ok(1));
     /// assert_eq!(zerovec.binary_search(&282), Err(2));
@@ -468,7 +551,17 @@ where
     #[inline]
     pub fn binary_search(&self, x: &T) -> Result<usize, usize> {
         self.as_slice()
-            .binary_search_by(|probe| T::from_unaligned(probe).cmp(x))
+            .binary_search_by(|probe| T::from_unaligned(*probe).cmp(x))
+    }
+}
+
+impl<T: AsULE> FromIterator<T> for ZeroVec<'_, T> {
+    /// Creates a [`ZeroVec::Owned`] from an iterator of values.
+    fn from_iter<I>(iter: I) -> Self
+    where
+        I: IntoIterator<Item = T>,
+    {
+        ZeroVec::Owned(iter.into_iter().map(|t| t.as_unaligned()).collect())
     }
 }
 
@@ -486,7 +579,7 @@ mod tests {
             assert_eq!(zerovec.get(2), Some(TEST_SLICE[2]));
         }
         {
-            let zerovec = ZeroVec::<u32>::try_from_bytes(TEST_BUFFER_LE).unwrap();
+            let zerovec = ZeroVec::<u32>::parse_byte_slice(TEST_BUFFER_LE).unwrap();
             assert_eq!(zerovec.get(0), Some(TEST_SLICE[0]));
             assert_eq!(zerovec.get(1), Some(TEST_SLICE[1]));
             assert_eq!(zerovec.get(2), Some(TEST_SLICE[2]));
@@ -501,7 +594,7 @@ mod tests {
             assert_eq!(Err(3), zerovec.binary_search(&0x0c0d0c));
         }
         {
-            let zerovec = ZeroVec::<u32>::try_from_bytes(TEST_BUFFER_LE).unwrap();
+            let zerovec = ZeroVec::<u32>::parse_byte_slice(TEST_BUFFER_LE).unwrap();
             assert_eq!(Ok(3), zerovec.binary_search(&0x0e0d0c));
             assert_eq!(Err(3), zerovec.binary_search(&0x0c0d0c));
         }
@@ -511,73 +604,74 @@ mod tests {
     fn test_odd_alignment() {
         assert_eq!(
             Some(0x020100),
-            ZeroVec::<u32>::try_from_bytes(TEST_BUFFER_LE)
+            ZeroVec::<u32>::parse_byte_slice(TEST_BUFFER_LE)
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x04000201),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[1..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[1..77])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x05040002),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[2..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[2..78])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x06050400),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[3..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[3..79])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x060504),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[4..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[4..])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x4e4d4c00),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[75..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[75..79])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x4e4d4c00),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[3..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[3..79])
                 .unwrap()
                 .get(18)
         );
         assert_eq!(
             Some(0x4e4d4c),
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[76..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[76..])
                 .unwrap()
                 .get(0)
         );
         assert_eq!(
             Some(0x4e4d4c),
-            ZeroVec::<u32>::try_from_bytes(TEST_BUFFER_LE)
+            ZeroVec::<u32>::parse_byte_slice(TEST_BUFFER_LE)
                 .unwrap()
                 .get(19)
         );
+        // TODO(#1144): Check for correct slice length in PlainOldULE
+        // assert_eq!(
+        //     None,
+        //     ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[77..])
+        //         .unwrap()
+        //         .get(0)
+        // );
         assert_eq!(
             None,
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[77..])
-                .unwrap()
-                .get(0)
-        );
-        assert_eq!(
-            None,
-            ZeroVec::<u32>::try_from_bytes(TEST_BUFFER_LE)
+            ZeroVec::<u32>::parse_byte_slice(TEST_BUFFER_LE)
                 .unwrap()
                 .get(20)
         );
         assert_eq!(
             None,
-            ZeroVec::<u32>::try_from_bytes(&TEST_BUFFER_LE[3..])
+            ZeroVec::<u32>::parse_byte_slice(&TEST_BUFFER_LE[3..79])
                 .unwrap()
                 .get(19)
         );
